/*
 * Copyright (c) 2014-2016 Alibaba Group. All rights reserved.
 * License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "lite-log_internal.h"


#if defined(LITE_LOG_ENABLED)
static log_client logcb;

static char *lvl_names[] = {
    "emg", "crt", "err", "wrn", "inf", "dbg",
};

void LITE_syslog(const char *f, const int l, const int level, const char *fmt, ...)
{
    char       *tmpbuf = logcb.text_buf;
    char       *o = tmpbuf;
    int         truncated = 0;
    va_list     ap;

    if (!strlen(LITE_get_logname()) || LITE_get_loglevel() < level || level < LOG_EMERG_LEVEL) {
        return;
    }

    LITE_printf(LOG_PREFIX_FMT, lvl_names[level], f, l);

    memset(tmpbuf, 0, sizeof(logcb.text_buf));
    va_start(ap, fmt);
    o += vsprintf(o, fmt, ap);
    va_end(ap);

    if (o - tmpbuf > LOG_MSG_MAXLEN) {
        o = tmpbuf + strlen(tmpbuf);
        truncated = 1;
    }
    if (strlen(tmpbuf) == LOG_MSG_MAXLEN) {
        truncated = 1;
    }

    LITE_printf("%s", tmpbuf);
    if (truncated) {
        LITE_printf(" ...");
    }

    LITE_printf("\r\n");
    return;
}

int LITE_log_enabled(void)
{
    return 1;
}

void LITE_openlog(const char *ident)
{
    strncpy(logcb.name, ident, LOG_MOD_NAME_LEN);
    logcb.name[LOG_MOD_NAME_LEN] = 0;
    logcb.priority = 0;
}

void LITE_closelog(void)
{
    strncpy(logcb.name, "", LOG_MOD_NAME_LEN);
    logcb.name[LOG_MOD_NAME_LEN] = 0;
    logcb.priority = 0;
}

char *LITE_get_logname(void)
{
    return logcb.name;
}

int LITE_get_loglevel(void)
{
    return logcb.priority;
}

void LITE_set_loglevel(int pri)
{
    logcb.priority = pri;
}

int log_multi_line_internal(const char *f, const int l,
                            const char *title, int level, char *payload, const char *mark)
{
    const char     *pos;
    const char     *endl;
    int             i;

    if (LITE_get_loglevel() < level) {
        return 1;
    }

    LITE_printf("[%s] %s(%d): %s (Length: %d Bytes)\r\n",
                lvl_names[LITE_get_loglevel()], f, l, title, (int)strlen(payload));

    pos = payload;
    while (pos && *pos) {
        LITE_printf("%s ", mark);

        if (*pos == '\r') {
            LITE_printf("\r\n");
            pos += 2;
            continue;
        }

        endl = strchr(pos + 1, '\r');
        if (endl == NULL) {
            endl = pos;
            do {
                ++endl;
            } while (*endl);
        }

        for (i = 0; i < endl - pos; ++i) {
            LITE_printf("%c", pos[i]);
        }
        LITE_printf("\r\n");

        pos = *endl ? endl + 2 : 0;
    }

    return 0;
}

#else   /* defined(LITE_LOG_ENABLED) */

void LITE_syslog(const char *f, const int l, const int level, const char *fmt, ...)
{
    va_list         ap;

    LITE_printf("%s/%d: ", f, l);
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);
    LITE_printf("\r\n");

    return;
}

int LITE_log_enabled(void)
{
    return 0;
}

void LITE_openlog(const char *ident)
{
    return;
}

void LITE_closelog(void)
{
    return;
}

char *LITE_get_logname(void)
{
    return NULL;
}

int LITE_get_loglevel(void)
{
    return -1;
}

void LITE_set_loglevel(int lvl)
{
    return;
}

int log_multi_line_internal(const char *f, const int l,
                            const char *title,
                            int level,
                            char *payload,
                            const char *mark)
{
    HAL_Printf("%s(%d):\r\n\r\n", f, l);
    HAL_Printf("%s\r\n", payload);

    return 0;
}

#endif  /* defined(LITE_LOG_ENABLED) */

